"""
This module contains the `Reaction` class representing a chemical process.

A reaction consists of one or more reactants and one or more products, all being
instances of the `StatefulSpecies` class (with or without states.).
A `Reaction` is instantiated by passing an ``r_str`` argument, a pyvalem-compatible
reaction string.

As with other pyvalem objects, the ``__repr__`` method is overloaded to provide a
*canonicalised* representation of the `Reaction` instance.

Examples
--------
>>> from pyvalem.reaction import Reaction
>>> Reaction('CO v=1 + O2 J=2;X(3SIGMA-g) → CO2 + O')
CO v=1 + O2 X(3Σ-g);J=2 → CO2 + O

>>> Reaction('CO + O2 <-> CO2 + O')
CO + O2 ⇌ CO2 + O

>>> Reaction('HCl + hv -> H+ + Cl + e-')
hν + HCl → H+ + Cl + e-

>>> str(Reaction('C6H5OH + 7O2 → 6CO2 + 3H2O'))
'C6H5OH + 7O2 → 6CO2 + 3H2O'

>>> repr(Reaction('C6H5OH + 7O2 → 6CO2 + 3H2O'))
'C6H5OH + O2 + O2 + O2 + O2 + O2 + O2 + O2 → CO2 + CO2 + CO2 + CO2 + CO2 + CO2 + H2O + H2O + H2O'
"""

import re

from .formula import FormulaParseError
from .stateful_species import StatefulSpecies


class ReactionParseError(Exception):
    pass


class ReactionStoichiometryError(ReactionParseError):
    pass


class ReactionChargeError(ReactionParseError):
    pass


class Reaction:
    """A class representing a reaction between species.

    A `Reaction` object is a collection of reactants and products, each of which is
    represented by a `StatefulSpecies` (though an individual species need not be
    associated with a specified state).
    It is instantiated using a string consisting of reactant and product
    `StatefulSpecies` valid strings, each separated by a plus sign surrounded by
    whitespace: ``" + "``.
    Reactants and products are separated by any of the following side separator strings:
    ``' → '``, ``' = '``, ``' ⇌ '``, ``' -> '``, ``' <-> '``, ``' <=> '``.

    The `Reaction` class implements ``html`` and ``latex`` attributes to render nice
    representations.

    Parameters
    ----------
    r_str : str
        PyValem-valid reaction string, see above.
    strict : bool, default=True
        If ``strict=False``, the stoichiometry and charge balance is not enforced.
        This is intended for incomplete or ambiguous reactions.

    Attributes
    ----------
    reactants : list of tuple[int, StatefulSpecies]
        Aggregated `StatefulSpecies` instances with their stoichiometries
    products : list of tuple[int, StatefulSpecies]
        Aggregated `StatefulSpecies` instances with their stoichiometries
    html
    latex

    Raises
    ------
    ReactionStoichiometryError
        If ``strict=True`` and the elemental stoichiometry is *not conserved*.
    ReactionChargeError
        If ``strict=True`` and the reaction charge is *not conserved*.

    Notes
    -----
    The ``__repr__`` method is overloaded to provie a *canonicalised* representation
    of the reaction. The idea is that two `Reaction` instances representing the same
    physical entity will have the same ``repr(reaction)`` text representation.
    The ``repr`` canonicalises the `StatefulSpecies` representations (all the way down
    to `Formula`, `State` and the states order), the sides separator, and expands all
    aggregated species (``"2H"`` becomes ``"H + H"``), while keeping the order of all
    the heavy species as passed as ``r_str``.
    All the light species, however, are aggregated and moved to the beginning of the
    *left-hand-side* of the reaction and to the end of the *right-hand-side* of the
    reaction.

    Examples
    --------
    >>> r1 = Reaction('CO v=1 + O2 J=2;X(3SIGMA-g) → CO2 + O')
    >>> r2 = Reaction('CO + O2 <-> CO2 + O')
    >>> r3 = Reaction('HCl + hv -> H+ + Cl + e-')
    >>> r4 = Reaction('C6H5OH + 7O2 → 6CO2 + 3H2O')

    >>> # Demonstration of the canonicalisation:
    >>> str(r3)  # non-canonical string repr.
    'HCl + hν → H+ + Cl + e-'

    >>> repr(r3)  # canonicalised representation
    'hν + HCl → H+ + Cl + e-'

    >>> str(r4)  # non-canonical string repr.
    'C6H5OH + 7O2 → 6CO2 + 3H2O'

    >>> repr(r4)  # canonicalised representation
    'C6H5OH + O2 + O2 + O2 + O2 + O2 + O2 + O2 → CO2 + CO2 + CO2 + CO2 + CO2 + CO2 + H2O + H2O + H2O'

    >>> # HTML attribute
    >>> r1.html
    'CO v=1 + O<sub>2</sub> J=2 X<sup>3</sup>Σ<sup>-</sup><sub>g</sub> → CO<sub>2</sub> + O'

    >>> # Consistency checks
    >>> Reaction('BeH+ + I2 <-> BeI + HI')
    Traceback (most recent call last):
      ...
    pyvalem.reaction.ReactionChargeError: Charge not preserved for reaction: BeH+ + I2 <-> BeI + HI

    >>> Reaction('BeH + I2 <-> BeI')
    Traceback (most recent call last):
      ...
    pyvalem.reaction.ReactionStoichiometryError: Stoichiometry not preserved for reaction: BeH + I2 <-> BeI

    >>> Reaction('BeH+ + I2 <->', strict=False)
    BeH+ + I2 ⇌
    """

    RP_SEPARATORS = "→", "=", "⇌", "->", "<->", "<=>"
    SPACED_RP_SEPARATORS = [" " + s + " " for s in RP_SEPARATORS]

    canonical_separators = {
        "→": "→",
        "->": "→",
        "=": "→",
        "⇌": "⇌",
        "<->": "⇌",
        "<=>": "⇌",
    }
    latex_sep = {"→": r"\rightarrow", "⇌": r"\rightlefthooks"}

    light_species = ("e-", "e+", "hv", "hν")

    def __init__(self, r_str, strict=True):
        # If the Reaction string has no products, add a space after the
        # separator (e.g. 'Ar + e- ->' becomes 'Ar + e- -> ').
        for sep in Reaction.RP_SEPARATORS:
            if r_str.rstrip().endswith(sep):
                r_str = r_str + " "
                break

        for sep in Reaction.SPACED_RP_SEPARATORS:
            fragments = r_str.split(sep)
            # fragments is a list: [LHS_str, RHS_str]
            if len(fragments) == 1:
                continue
            elif len(fragments) > 2:
                raise ReactionParseError(
                    "Invalid reaction string - multiple reactant-product "
                    "separators: {}".format(r_str)
                )
            break
        else:
            raise ReactionParseError(
                "Invalid reaction string - no reactant-product separator: "
                "{}".format(r_str)
            )

        # canonicalised side separator:
        self.sep = self.canonical_separators[sep.strip()]

        # parse reactants and products into
        # dict[str, int]
        # and
        # list[tuple[int: stoichiometry, StatefulSpecies]]
        self.reactants_text_count_map = None
        self.products_text_count_map = None
        self.reactants = None
        self.products = None
        try:
            self._parse_species(*fragments)
        except FormulaParseError as err:
            raise ReactionParseError(
                'Failed to parse Reaction string "{}" because one of the '
                "StatefulSpecies was incorrectly formed. "
                "The error reported was: {}".format(r_str, err)
            )

        # validate charge and stoichiometry conservation:
        if strict and not self.stoichiometry_conserved():
            raise ReactionStoichiometryError(
                "Stoichiometry not preserved for " "reaction: {}".format(r_str)
            )
        if strict and not self.charge_conserved():
            raise ReactionChargeError(
                "Charge not preserved for " "reaction: {}".format(r_str)
            )

    def _parse_species(self, lhs_str, rhs_str):
        """Parse strings of both sides of the reaction string

        Parses into self.reactants and self.products in the form of lists of
        (species_count, StatefulSpecies) tuples.
        On the side, also creates maps between species string (as passed
        to the constructor) and number of species on the side.
        Those are self.reactants_text_count_map and self.products_text_count_map.

        Populates ``self.reactants``, ``self.products``,
        ``self.reactants_text_count_map``, and ``self.products_text_count_map``

        Parameters
        ----------
        lhs_str, rhs_str : str
            Strings of reactants and products sides, around the sides separator from
            the ``r_str``.
        """
        self.reactants_text_count_map = {}
        self.products_text_count_map = {}
        self.reactants = []
        self.products = []
        for side_str, species, species_map in zip(
            [lhs_str, rhs_str],
            [self.reactants, self.products],
            [self.reactants_text_count_map, self.products_text_count_map],
        ):
            for side_fragment in side_str.split(" + "):
                if side_fragment.strip():
                    patt = r"(\d*)(.*)"
                    n, ss_str = re.match(patt, side_fragment).groups()
                    if not n:
                        n = 1
                    else:
                        try:
                            n = int(n)
                        except ValueError:
                            raise ReactionParseError(
                                "Failed to parse {}".format(side_fragment)
                            )
                    ss = StatefulSpecies(ss_str)
                    species.append((n, ss))
                    ss_str = str(ss)
                    if ss_str not in species_map:
                        species_map[ss_str] = 0
                    species_map[ss_str] += n

    @staticmethod
    def _sort_terms(terms, side="lhs"):
        """Sorts the species for the canonicalised representation.

        The only sorting is performed on light species, these are placed
        first on LHS and last on RHS of the reaction.

        Parameters
        ----------
        terms : iterable of tuple[int, StatefulSpecies]
        side : {"lhs", "rhs"}, default="lhs"

        Returns
        -------
        list of tuple[int, StatefulSpecies]
        """
        sort_keys = {sp: (-1, sp) for sp in Reaction.light_species}
        return list(
            sorted(
                terms,
                key=lambda e: sort_keys.get(e[1].formula.formula, (0, "")),
                reverse=side != "lhs",
            )
        )

    def _aggregate_terms(self, terms):
        """Aggregates the same light species into a single term.

        Accepts list[tuple[int, StatefulSpecies]].
        Collect separate terms involving the same StatefulSpecies into a single
        term with the same total stoichiometry.
        The aggregation is only done on successive light species terms. This is
        connected to the canonicalisation by the ``__repr__`` method.

        Parameters
        ----------
        terms : list of tuple[int, StatefulSpecies]

        Returns
        -------
        list of tuple[int, StatefulSpecies]

        Examples
        --------
        H + H + e + e + He + e + e -> H + H + 2e + He + 2e
        """
        if not len(terms):
            return []
        aggregated_terms = [terms[0]]
        for term in terms[1:]:
            if term[1].formula.formula in self.light_species:
                last_term = aggregated_terms[-1]
                if term[1] == last_term[1]:
                    n = term[0] + last_term[0]
                    aggregated_terms[-1] = (n, term[1])
                else:
                    aggregated_terms.append(term)
            else:
                aggregated_terms.append(term)
        return aggregated_terms

    def _expand_terms(self, terms):
        """Method to expand all aggregated heavy species terms.

        Used for the canonicalised representation given by ``__repr__``.

        Parameters
        ----------
        terms : list of tuple[int, StatefulSpecies]

        Returns
        -------
        list of tuple[int, StatefulSpecies]
        """
        expanded_terms = []
        for n, ss in terms:
            if ss.formula.formula in self.light_species:
                expanded_terms.append((n, ss))
            else:
                for _ in range(n):
                    expanded_terms.append((1, ss))
        return expanded_terms

    @staticmethod
    def _silent_n(n):
        return str(n) if n != 1 else ""

    def __str__(self):
        reactants_str = " + ".join(
            "{}{}".format(self._silent_n(n), str(ss)) for n, ss in self.reactants
        )
        products_str = " + ".join(
            "{}{}".format(self._silent_n(n), str(ss)) for n, ss in self.products
        )
        return "{} {} {}".format(reactants_str, self.sep, products_str).strip()

    def __repr__(self):
        """
        Performs canonicalisation of the reaction string by expanding
        aggregated stoichiometries of all the heavy species and by aggregating
        light species (e, hv) and moving them to the side.
        """
        reactants = self._sort_terms(self.reactants, side="lhs")
        products = self._sort_terms(self.products, side="rhs")
        reactants = self._aggregate_terms(reactants)
        products = self._aggregate_terms(products)
        reactants = self._expand_terms(reactants)
        products = self._expand_terms(products)

        reactants_repr = " + ".join(
            "{}{}".format(self._silent_n(n), repr(ss)) for n, ss in reactants
        )
        products_repr = " + ".join(
            "{}{}".format(self._silent_n(n), repr(ss)) for n, ss in products
        )
        return "{} {} {}".format(reactants_repr, self.sep, products_repr).strip()

    @staticmethod
    def _get_all_stoichs(side):
        stoich = {}
        for n, ss in side:
            for sp, nn in ss.formula.atom_stoich.items():
                if sp in stoich:
                    stoich[sp] += n * nn
                else:
                    stoich[sp] = n * nn
        return stoich

    def stoichiometry_conserved(self):
        """Verify that the `Reaction` instance conserves its stoichiometry.

        Returns
        -------
        bool
        """
        reactants_stoich = self._get_all_stoichs(self.reactants)
        products_stoich = self._get_all_stoichs(self.products)

        return reactants_stoich == products_stoich

    @staticmethod
    def _get_total_charge(side):
        return sum(n * ss.formula.charge for n, ss in side if ss.formula.formula != "M")

    def charge_conserved(self):
        """Verify that the `Reaction` instance conserves its charge.

        Returns
        -------
        bool
        """
        reactants_total_charge = self._get_total_charge(self.reactants)
        products_total_charge = self._get_total_charge(self.products)

        return reactants_total_charge == products_total_charge

    def __eq__(self, other):
        return repr(self) == repr(other)

    @property
    def html(self):
        """HTML representation of the `Reaction` instance.

        Returns
        -------
        str
        """
        reactants_html = " + ".join(
            "{}{}".format(self._silent_n(n), ss.html) for n, ss in self.reactants
        )
        products_html = " + ".join(
            "{}{}".format(self._silent_n(n), ss.html) for n, ss in self.products
        )
        return "{} {} {}".format(reactants_html, self.sep, products_html).strip()

    @property
    def latex(self):
        """LaTeX representation of the `Reaction` instance.

        Returns
        -------
        str
        """
        reactants_latex = " + ".join(
            "{}{}".format(self._silent_n(n), ss.latex) for n, ss in self.reactants
        )
        products_latex = " + ".join(
            "{}{}".format(self._silent_n(n), ss.latex) for n, ss in self.products
        )
        return "{} {} {}".format(
            reactants_latex, self.latex_sep[self.sep], products_latex
        ).strip()



# import all the State classes and exceptions into the states namespace:
from ._base_state import State, StateParseError
from .atomic_configuration import AtomicConfiguration, AtomicConfigurationError
from .diatomic_molecular_configuration import (
    DiatomicMolecularConfiguration,
    DiatomicMolecularConfigurationError,
)
from .atomic_term_symbol import AtomicTermSymbol, AtomicTermSymbolError
from .molecular_term_symbol import MolecularTermSymbol, MolecularTermSymbolError
from .racah_symbol import RacahSymbol
from .rotational_state import RotationalState, RotationalStateError
from .vibrational_state import VibrationalState, VibrationalStateError
from .generic_excited_state import GenericExcitedState, GenericExcitedStateError
from .key_value_pair import KeyValuePair, KeyValuePairError



"""
The MolecularTermSymbol class, representing a molecular term symbol, with
methods for parsing a string into quantum numbers and labels, creating
an HTML representation of the term symbol, etc.
"""
import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError
from pyvalem._utils import parse_fraction, float_to_fraction

orbital_irrep_labels = (
    "Σ-",
    "Σ-u",
    "Σ-g",
    "SIGMA-",
    "SIGMA-u",
    "SIGMA-g",
    "Σ+",
    "Σ+u",
    "Σ+g",
    "SIGMA+",
    "SIGMA+u",
    "SIGMA+g",
    "Π",
    "Πu",
    "Πg",
    "PI",
    "PIu",
    "PIg",
    "Δ",
    "Δu",
    "Δg",
    "DELTA",
    "DELTAu",
    "DELTAg",
    "Φ",
    "Φg",
    "Φu",
    "PHI",
    "PHIg",
    "PHIu",
    "Γ",
    "Γg",
    "Γu",
    "GAMMA",
    "GAMMAg",
    "GAMMAu",
    "A",
    "A'",
    'A"',
    "A1",
    "A'1",
    'A"1',
    "A2",
    "A'2",
    'A"2',
    "A3",
    "Ag",
    "A'g",
    'A"g',
    "A1g",
    "A2g",
    "A3g",
    "Au",
    "A'u",
    'A"u',
    "A1u",
    "A2u",
    "A3u",
    "B",
    "B'",
    'B"',
    "B1",
    "B2",
    "B3",
    "Bg",
    "B'g",
    'B"g',
    "B1g",
    "B2g",
    "B3g",
    "Bu",
    "B'u",
    'B"u',
    "B1u",
    "B2u",
    "B3u",
    "E",
    "E'",
    'E"',
    "E1",
    "E'1",
    'E"1',
    "E2",
    "E'2",
    'E"2',
    "E3",
    "E'3",
    'E"3',
    "E4",
    "E5",
    "E6",
    "E7",
    "Eg",
    "E'g",
    'E"g',
    "E1g",
    "E2g",
    "E3g",
    "Eu",
    "E'u",
    'E"u',
    "E1u",
    "E2u",
    "E3u",
    "T",
    "T1",
    "T2",
    "Tg",
    "T1g",
    "T2g",
    "Tu",
    "T1u",
    "T2u",
    "G",
    "Gg",
    "Gu",
    "H",
    "Hg",
    "Hu",
)

greek_letters = {
    "SIGMA-": "Σ-",
    "SIGMA-u": "Σ-u",
    "SIGMA-g": "Σ-g",
    "SIGMA+": "Σ+",
    "SIGMA+u": "Σ+u",
    "SIGMA+g": "Σ+g",
    "PI": "Π",
    "PIu": "Πu",
    "PIg": "Πg",
    "DELTA": "Δ",
    "DELTAu": "Δu",
    "DELTAg": "Δg",
    "PHI": "Φ",
    "PHIg": "Φg",
    "PHIu": "Φu",
    "GAMMA": "Γ",
    "GAMMAg": "Γg",
    "GAMMAu": "Γu",
}

latex_terms = {
    "Σ": r"\Sigma",
    "Π": r"\Pi",
    "Δ": r"\Delta",
    "Φ": r"\Phi",
    "Γ": r"\Gamma",
}

integer = pp.Word(pp.nums)
molecule_Smult = integer.setResultsName("Smult")
molecule_irrep = pp.oneOf(orbital_irrep_labels).setResultsName("irrep")
molecule_Omegastr = (
    pp.Combine(pp.Optional(pp.oneOf(("+", "-"))) + integer)
    + pp.Optional(pp.Suppress("/") + "2")
).setResultsName("Omegastr")
molecule_term = (
    molecule_Smult + molecule_irrep + pp.Optional(pp.Suppress("_") + molecule_Omegastr)
)

# Term Symbol label can be a single letter (perhaps followed by ' or ") ...
term_label_let = pp.Combine(
    pp.Word(pp.srange("[A-Za-z]")) + pp.Optional(pp.oneOf(("'", '"')))
).setResultsName("term_label")
# ... or a number. The open bracket is required but suppressed to distinguish
# this type of term label from the spin multiplicity.
term_label_num = integer.setResultsName("term_label") + pp.Suppress("(")
term_label = term_label_let | term_label_num

molecule_term_with_label = (
    pp.Optional(term_label)
    + pp.Suppress(pp.Optional("("))
    + molecule_term
    + pp.Suppress(pp.Optional(")"))
    + pp.StringEnd()
)


class MolecularTermSymbolError(StateParseError):
    pass


class MolecularTermSymbol(State):
    def __init__(self, state_str):
        self.state_str = state_str
        self.Smult = None
        self.S = None
        self.irrep = None
        self.term_label = None
        self.Omega = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):

        try:
            components = molecule_term_with_label.parseString(state_str)
        except pp.ParseException:
            raise MolecularTermSymbolError(
                "Invalid molecular term symbol syntax: {0}".format(state_str)
            )
        self.Smult = int(components.Smult)
        self.S = (self.Smult - 1) / 2
        irrep = components.irrep
        if irrep in greek_letters.keys():
            self.state_str = self.state_str.replace(irrep, greek_letters[irrep])
            irrep = greek_letters[irrep]

        self.irrep = irrep
        self.term_label = components.term_label or None
        self.Omega = parse_fraction(components.Omegastr)

    @staticmethod
    def _irrep_html(irrep):
        irrep_chunks = []
        if "+" in irrep or "-" in irrep:
            irrep_chunks.append("{0:s}<sup>{1:s}</sup>".format(irrep[0], irrep[1]))
            next_idx = 2
        elif '"' in irrep or "'" in irrep:
            irrep_chunks.append(irrep[0:2])
            next_idx = 2
        else:
            irrep_chunks.append(irrep[0])
            next_idx = 1
        if irrep[next_idx:] != "":
            irrep_chunks.append("<sub>{:s}</sub>".format(irrep[next_idx:]))
        return "".join(irrep_chunks)

    @staticmethod
    def _irrep_latex(irrep):
        irrep_chunks = []
        if "+" in irrep or "-" in irrep:
            irrep_chunks.append(r"\Sigma^{:s}".format(irrep[1]))
            next_idx = 2
        elif '"' in irrep:
            irrep_chunks.append(irrep[0] + "''")
            next_idx = 2
        elif "'" in irrep:
            irrep_chunks.append(irrep[0:2])
            next_idx = 2
        else:
            symbol = irrep[0]
            symbol = latex_terms.get(irrep[0]) or symbol
            irrep_chunks.append(symbol)
            next_idx = 1
        if irrep[next_idx:] != "":
            irrep_chunks.append("_{{{:s}}}".format(irrep[next_idx:]))
        return "".join(irrep_chunks)

    @property
    def html(self):
        """Return an HTML representation of the MolecularTermSymbol."""

        # TODO Handle terms with superscripts and subscripts more gracefully.

        html_chunks = []
        if self.term_label is not None:
            html_chunks.append("{:s}".format(self.term_label))
        html_chunks.append(
            "<sup>{0:d}</sup>{1:s}".format(self.Smult, self._irrep_html(self.irrep))
        )
        if self.Omega is not None:
            omegastr = float_to_fraction(self.Omega)
            html_chunks.append("<sub>{0:s}</sub>".format(omegastr))
        return "".join(html_chunks)

    @property
    def latex(self):
        """Return an LaTeX representation of the MolecularTermSymbol."""

        latex_chunks = []
        if self.term_label is not None:
            latex_chunks.append(self.term_label)
        latex_chunks.append(
            "{{}}^{{{:d}}}{:s}".format(self.Smult, self._irrep_latex(self.irrep))
        )
        if self.Omega is not None:
            omegastr = float_to_fraction(self.Omega)
            # NB avoid double underscore if we have a subscript already.
            latex_chunks.append("{{}}_{{{:s}}}".format(omegastr))

        return "".join(latex_chunks)

    def __repr__(self):
        return self.state_str



"""
Special cases of formulas for particular species.

The keys of the inner dictionaries are attribute names of the resulting Formula
class instance, and are assigned to the corresponding values.
"""

special_cases = {
    "M": {
        "slug": "M",
        "html": "M",
        "latex": r"\mathrm{M}",
        "atom_stoich": {},
        "mass": None,
        "rmm": None,
        "charge": None,
        "natoms": None,
        "atoms": {"M"},
        "name": "generic third body",
    },
    "hν": {
        "slug": "hv",
        "html": "hν",
        "latex": r"h\nu",
        "atom_stoich": {},
        "mass": 0,
        "rmm": 0,
        "charge": 0,
        "natoms": None,
        "atoms": {},
        "name": "photon",
    },
    "e-": {
        "slug": "e_m",
        "html": "e<sup>-</sup>",
        "latex": r"\mathrm{e}^-",
        "atom_stoich": {},
        "mass": 5.48579909e-04,  # m_e / u
        "rmm": 5.48579909e-04,  # m_e / u
        "charge": -1,
        "natoms": None,
        "atoms": {},
        "name": "electron",
    },
    "e+": {
        "slug": "e_p",
        "html": "e<sup>+</sup>",
        "latex": r"\mathrm{e}^+",
        "atom_stoich": {},
        "mass": 5.48579909e-04,  # m_(e+) / u
        "rmm": 5.48579909e-04,  # m_(e+) / u
        "charge": 1,
        "natoms": None,
        "atoms": {},
        "name": "positron",
    },
}
special_cases["hv"] = special_cases["hν"]



"""
The DiatomicMolecularConfiguration class, representing a diatomic molecular
configuration with methods for parsing a string into quantum numbers and
labels, creating an HTML representation of the term symbol, etc.
"""

import pyparsing as pp

from .atomic_configuration import atomic_orbital_symbols
from pyvalem.states._base_state import State, StateParseError

integer = pp.Word(pp.nums)
molecular_orbital_symbols = (
    "σ",
    "π",
    "δ",
    "sigma",
    "pi",
    "delta",
    "σg",
    "πg",
    "δg",
    "σu",
    "πu",
    "δu",
    "sigmag",
    "pig",
    "deltag",
    "sigmau",
    "piu",
    "deltau",
)

greek_letters = {
    "sigma": "σ",
    "pi": "π",
    "delta": "δ",
    "sigmag": "σg",
    "sigmau": "σu",
    "pig": "πg",
    "piu": "πu",
    "deltag": "δg",
    "deltau": "δu",
}
symbol_latex = {"σ": r"\sigma", "π": r"\pi", "δ": r"\delta"}

molecule_orbital = pp.Group(
    integer.setResultsName("n")
    + pp.oneOf(molecular_orbital_symbols).setResultsName("symbol")
    + pp.Optional(integer.setResultsName("count"), default="1")
)

molecule_config = (
    molecule_orbital
    + pp.ZeroOrMore(pp.Suppress(".") + molecule_orbital)
    + pp.StringEnd()
).leaveWhitespace()

# atomic_orbital = pp.Group(integer.setResultsName('n') +
#                    pp.oneOf(atomic_orbital_symbols).setResultsName('lletter')
#                         )
alt_molecule_orbital = pp.Group(
    integer.setResultsName("n")
    + pp.oneOf(atomic_orbital_symbols).setResultsName("lletter")
    + pp.Suppress("-")
    + pp.oneOf(molecular_orbital_symbols).setResultsName("symbol")
    + pp.Optional(integer.setResultsName("count"), default="1")
)
alt_molecule_config = (
    alt_molecule_orbital
    + pp.ZeroOrMore(pp.Suppress(".") + alt_molecule_orbital)
    + pp.StringEnd()
).leaveWhitespace()


class DiatomicMolecularOrbitalError(StateParseError):
    pass


class DiatomicMolecularConfigurationError(StateParseError):
    pass


class DiatomicMolecularOrbital:
    def __init__(self, n, symbol, count):
        self.n = int(n)
        if symbol in greek_letters.keys():
            self.symbol = greek_letters[symbol]
        else:
            self.symbol = symbol
        self.count = int(count)
        self.validate_molecular_orbital()

    def __repr__(self):
        return "{:d}{:s}{:d}".format(self.n, self.symbol, self.count)

    def __hash__(self):
        return hash((self.n, self.symbol, self.count))

    def __eq__(self, other):
        return (
            self.n == other.n
            and self.symbol == other.symbol
            and self.count == other.count
        )

    def validate_molecular_orbital(self):
        if self.symbol == "σ":
            if self.count > 2:
                raise DiatomicMolecularOrbitalError(
                    "Only two electrons"
                    " allowed in σ orbital, but received {:d}".format(self.count)
                )
        else:
            if self.count > 4:
                raise DiatomicMolecularOrbitalError(
                    "Only four electrons"
                    " allowed in π and δ orbitals, "
                    "but received {:d}".format(self.count)
                )

    @staticmethod
    def _make_symbol_html(symbol):
        if symbol[-1] in "ug":
            return symbol[:-1] + "<sub>{}</sub>".format(symbol[-1])
        return symbol

    @property
    def html(self):
        return "{:d}{:s}<sup>{:d}</sup>".format(
            self.n, self._make_symbol_html(self.symbol), self.count
        )

    @staticmethod
    def _make_symbol_latex(symbol):
        if symbol[-1] in "ug":
            return symbol_latex[symbol[:-1]] + "_{}".format(symbol[-1])
        return symbol_latex[symbol]

    @property
    def latex(self):
        return "{:d}{:s}^{{{:d}}}".format(
            self.n, self._make_symbol_latex(self.symbol), self.count
        )


class AltDiatomicMolecularOrbital(DiatomicMolecularOrbital):
    """
    Alternative syntax for diatomic molecular orbitals: e.g. 1s-σg2.1s-σu1

    """

    def __init__(self, n, lletter, symbol, count):
        self.lletter = lletter
        super().__init__(n, symbol, count)

    def __repr__(self):
        return "{}{}-{}{}".format(self.n, self.lletter, self.symbol, self.count)

    def __hash__(self):
        return hash((self.n, self.lletter, self.symbol, self.count))

    def __eq__(self, other):
        return (
            self.n == self.n
            and self.lletter == self.lletter
            and self.symbol == self.symbol
            and self.count == self.count
        )

    @property
    def html(self):
        return "{:d}{}{:s}<sup>{:d}</sup>".format(
            self.n, self.lletter, self._make_symbol_html(self.symbol), self.count
        )

    @property
    def latex(self):
        return "{:d}{}{:s}^{{{:d}}}".format(
            self.n, self.lletter, self._make_symbol_latex(self.symbol), self.count
        )


class DiatomicMolecularConfiguration(State):
    def __init__(self, state_str):
        self.state_str = state_str
        self.orbitals = []
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        if "-" in state_str:
            self._parse_alt_config(state_str)
        else:
            self._parse_regular_config(state_str)

    def _parse_regular_config(self, state_str):
        try:
            parse_results = molecule_config.parseString(state_str)
        except pp.ParseException:
            raise DiatomicMolecularConfigurationError(
                "Invalid diatomic molecular electronic "
                "configuration syntax: {0}".format(state_str)
            )

        for i, parsed_orbital in enumerate(parse_results):
            try:
                temp_orbital = DiatomicMolecularOrbital(
                    n=parsed_orbital["n"],
                    symbol=parsed_orbital["symbol"],
                    count=parsed_orbital["count"],
                )
            except DiatomicMolecularOrbitalError as err:
                raise DiatomicMolecularConfigurationError(err)
            self.orbitals.append(temp_orbital)
        self._validate_configuration()

    def _parse_alt_config(self, state_str):
        try:
            parse_results = alt_molecule_config.parseString(state_str)
        except pp.ParseException:
            raise DiatomicMolecularConfigurationError(
                "Invalid diatomic molecular electronic "
                "configuration syntax: {0}".format(state_str)
            )

        self.orbitals = []
        for i, parsed_orbital in enumerate(parse_results):
            try:
                temp_orbital = AltDiatomicMolecularOrbital(
                    n=parsed_orbital["n"],
                    lletter=parsed_orbital["lletter"],
                    symbol=parsed_orbital["symbol"],
                    count=parsed_orbital["count"],
                )
            except DiatomicMolecularOrbitalError as err:
                raise DiatomicMolecularConfigurationError(err)
            self.orbitals.append(temp_orbital)
        self._validate_alt_configuration()

    def _validate_configuration(self):
        orbitals = [(o.n, o.symbol) for o in self.orbitals]
        if len(orbitals) != len(set(orbitals)):
            raise DiatomicMolecularConfigurationError(
                "Repeated orbitals " "in {0}".format(self.state_str)
            )

    def _validate_alt_configuration(self):
        orbitals = [(o.n, o.lletter, o.symbol) for o in self.orbitals]
        if len(orbitals) != len(set(orbitals)):
            raise DiatomicMolecularConfigurationError(
                "Repeated orbitals " "in {0}".format(self.state_str)
            )

    def __repr__(self):
        return ".".join([str(orbital) for orbital in self.orbitals])

    def __hash__(self):
        return hash(self.state_str)

    @property
    def html(self):
        html_chunks = []
        if self.orbitals:
            for orbital in self.orbitals:
                html_chunks.append(orbital.html)
        return ".".join(html_chunks)

    @property
    def latex(self):
        latex_chunks = []
        if self.orbitals:
            for orbital in self.orbitals:
                latex_chunks.append(orbital.latex)
        return "".join(latex_chunks)

    def __eq__(self, other):
        return self.orbitals == other.orbitals



"""
Utility functions for the pyvalem package.
"""


def parse_fraction(s):
    """Parse an integer or fraction as a pyparsing.ParseResults object resulting
    from a match to '2', '3/2', etc. into a float.

    If s is None or the empty string, returns None.

    Parameters
    ----------
    s : str or None

    Returns
    -------
    float or None
    """
    if not s:
        return None
    if len(s) == 2:
        num, det = s
        return float(num) / int(det)
    elif len(s) == 1:
        if s:
            return float(s[0])

    raise ValueError(s, "does not parse to an integer or fraction")


def float_to_fraction(v):
    """Convert the integer or half-integer float v into a string fraction.

    Parameters
    ----------
    v : number

    Returns
    -------
    str
    """

    if v.is_integer():
        return str(int(v))
    return "{0:d}/2".format(int(2 * v))



"""
The GenericExcitedState class, representing an unspecified excited state of
an atom, ion or molecule as '*', '**', '***', '****', '5*', etc.
"""

import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError

integer = pp.Word(pp.nums)
atom_int = integer.setResultsName("int")
state_term = (atom_int + "*" + pp.StringEnd()).leaveWhitespace()


class GenericExcitedStateError(StateParseError):
    pass


class GenericExcitedState(State):
    def __init__(self, state_str):
        self.state_str = state_str
        self.int_n = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        if "*" in state_str:
            if state_str.count("*") == len(state_str):
                if len(state_str) > 4:
                    raise StateParseError(
                        "Invalid excited state value: {0}"
                        " Can be *, **, ***, or ****".format(state_str)
                    )
            else:
                try:
                    components = state_term.parseString(state_str)
                    self.int_n = int(components.int)
                except pp.ParseException:
                    raise GenericExcitedStateError(
                        "Invalid excited state value syntax: {0} has to be of form"
                        " n* with n = integer".format(state_str)
                    )
        else:
            raise GenericExcitedStateError(
                "Invalid excited state value"
                " syntax: {0} Must have a * term.".format(state_str)
            )

    def __repr__(self):
        if self.state_str == "1*":
            return "*"
        elif self.state_str == "2*":
            return "**"
        elif self.state_str == "***":
            return "3*"
        elif self.state_str == "****":
            return "4*"
        return self.state_str

    @property
    def html(self):
        return "<sup>{:s}</sup>".format(repr(self))

    @property
    def latex(self):
        return "^{{{:s}}}".format(repr(self))



"""
A State class, representing a quantum state or label of a species.

This is an abstract base class and specific types of state derive from it: use
one of those and don't instantiate State objects directly.
"""

from abc import ABC, abstractmethod
import html


class StateError(Exception):
    """A base class for state related exceptions."""

    pass


class StateParseError(StateError):
    """A base class for state parsing related exceptions."""

    pass


# noinspection PyUnresolvedReferences
class State(ABC):
    multiple_allowed = False

    @property
    def html(self):
        """HTML representation of the State instance.

        Returns
        -------
        str
        """
        return html.escape(repr(self))

    @property
    def latex(self):
        """LaTeX representation of the State instance.

        Returns
        -------
        str
        """
        return repr(self)

    @abstractmethod
    def __repr__(self):
        """In the context of PyValem package, the __repr__ method is overloaded to
        provide a *canonicalised* string representation of the structure. The
        main idea is that two objects which represent *the same* physical entity will
        have the same ``repr(obj)`` representation.
        """
        raise NotImplementedError

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.__repr__() == other.__repr__():
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__repr__())



"""
A module for parsing strings or sequences of strings into appropriate
State-like objects or sequences of such objects.
"""
from collections import OrderedDict

from ._base_state import StateParseError
from .atomic_configuration import AtomicConfiguration
from .atomic_term_symbol import AtomicTermSymbol
from .diatomic_molecular_configuration import DiatomicMolecularConfiguration
from .generic_excited_state import GenericExcitedState
from .key_value_pair import KeyValuePair
from .molecular_term_symbol import MolecularTermSymbol
from .racah_symbol import RacahSymbol
from .rotational_state import RotationalState
from .vibrational_state import VibrationalState

# the following has two purposes: keys determine the order in which the
# states are parsed, and the values determine the sorting order of states
# for StatefulSpecies.__repr__.
STATES = OrderedDict(
    [
        (GenericExcitedState, 0),
        (AtomicConfiguration, 1),
        (AtomicTermSymbol, 2),
        (DiatomicMolecularConfiguration, 1),
        (MolecularTermSymbol, 2),
        (VibrationalState, 3),
        (RotationalState, 4),
        (RacahSymbol, 5),
        (KeyValuePair, 6),
    ]
)


def state_parser(s_state):
    """Parse s_state into an appropriate State-like object or list of such."""

    if not s_state:
        return None

    if not isinstance(s_state, str):
        # If we have a sequence of strings, parse them one by one into a list
        # of State-like objects.
        return [state_parser(s.strip()) for s in s_state]

    # Try to parse the string s_state into a State-like object by trying each
    # of the possible derived State classes one by one in a particular order.
    for StateClass in STATES:
        try:
            return StateClass(s_state)
        except StateParseError:
            pass

    raise StateParseError("Could not parse {}".format(s_state))



"""
The RacahSymbol class, for the representation of an atomic state in Racah
notation, with methods for parsing a string into quantum numbers and
labels, creating an HTML representation of the term symbol, etc.
"""

import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError

integer = pp.Word(pp.nums)
atom_principal = integer.setResultsName("principal")

orbital_labels = ("s", "s'", "p", "p'", "d", "d'", "f", "f'")
atom_orbital = pp.oneOf(orbital_labels).setResultsName("orbital")

atom_k_term = (
    integer.setResultsName("k_num") + pp.Suppress("/") + integer.setResultsName("k_den")
)

atom_j_term = integer.setResultsName("jterm")

racah_symbol_template = (
    atom_principal
    + atom_orbital
    + pp.Suppress("[")
    + atom_k_term
    + pp.Suppress("]_")
    + atom_j_term
)


class RacahSymbol(State):
    """
    The RacahSymbol class representing an atomic state in Racah notation.

    This class is currently little more than a stub. The term symbol of the
    parent configuration must be specified as a separate State; this class
    defines nl[K]_J, without the parity label ("superscript-o").
    """

    def __init__(self, state_str):
        self.state_str = state_str
        self.principal = None
        self.orbital = None
        self.parent_rot = None
        self.k_num = None
        self.k_den = None
        self.j_term = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        try:
            components = racah_symbol_template.parseString(state_str)
        except pp.ParseException:
            raise StateParseError("Invalid Racah notation syntax: {}".format(state_str))
        self.principal = int(components.principal)
        self.orbital = components.orbital
        if "'" in self.orbital:
            self.parent_rot = 0.5
        else:
            self.parent_rot = 1.5
        self.k_num = int(components.k_num)
        self.k_den = int(components.k_den)
        self.j_term = int(components.jterm)

    def __repr__(self):
        parent = "{}{}".format(self.principal, self.orbital)
        k = "{}/{}".format(self.k_num, self.k_den)
        return "{}[{}]_{}".format(parent, k, self.j_term)

    @property
    def html(self):
        parent = "{}{}".format(self.principal, self.orbital)
        k = "{}/{}".format(self.k_num, self.k_den)
        return "{}[{}]<sub>{}</sub>".format(parent, k, self.j_term)



"""This module defines dictionaries of meta data relating to the elements and
their isotopes.

Routine Listings
----------------
atoms : dict[str, Atom]
    Dictionary of `Atom` instances keyed by the elemental symbols.
isotopes : dict[str, Isotope]:
    Dictionary of `Isotope` instances. Keys are in the form of mass number followed by
    the element symbol, such as ``"38Ar"``

Notes
-----
The atomic data are compiled from Meija et al.[1]_.
The isotope data are compiled from the AME2016 Atomic Mass Evaluation reports[2]_.

References
----------
.. [1] Meija et al., "Atomic weights of the elements 2013
   (IUPAC Technical Report)", Pure Appl. Chem. 88(3), 265-291, 2016.
   See https://ciaaw.org/atomic-weights.htm
.. [2] Huang et al., "The Ame2016 atomic mass evaluation (I)", Chinese Physics C41,
   030002 (2017); Wang et al., "The Ame2016 atomic mass evaluation (II)",
   Chinese Physics C41, 030003 (2017).
   See http://amdc.impcas.ac.cn/masstables/Ame2016/mass16.txt

Examples
--------
>>> ar_atom = atoms["Ar"]
>>> ar_atom.name
'Argon'
>>> ar_atom.Z
18

>>> ar_isotope = isotopes["38Ar"]
>>> ar_isotope.A, ar_isotope.N
(38, 20)
"""

import csv

try:
    import importlib.resources as pkg_resources
except ImportError:
    # for python < 3.7, use the importlib-resources backport
    # noinspection PyUnresolvedReferences
    import importlib_resources as pkg_resources


class Atom:
    """Class representing an atom instance.

    All the parameters are stored as instance attributes.
    The `Atom` instances are hashable.

    Parameters
    ----------
    symbol, name : str
        Element symbol and name, such as ``"Ar"``, ``"Argon"``.
    atomic_number : int
        Z number of the atom.
    weight, weight_unc : float, optional
        Atomic weight in [amu] and the uncertainty.

    Attributes
    ----------
    symbol, name : str
    Z : int
    weight, weight_unc : float or NoneType

    Examples
    --------
    >>> atom = Atom("Ar", "Argon", 18)
    >>> atom == "Ar"
    True
    >>> atom.is_isotope
    False
    """

    is_isotope = False

    def __init__(self, symbol, name, atomic_number, weight=None, weight_unc=None):
        self.symbol = symbol
        self.name = name
        self.Z = atomic_number
        self.weight = self.mass = weight
        self.weight_unc = weight_unc

    def __repr__(self):
        return self.symbol

    def __hash__(self):
        return hash(self.symbol)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.symbol == other.symbol
        return self.symbol == other


class Isotope(Atom):
    """Class representing an isotope instance.

    The `Isotope` instances are hashable.

    Parameters
    ----------
    atomic_number : int
        Atomic number Z of the isotope.
    mass_number : int
        Mass number A of the isotope.
    symbol, name : str
        Element symbol and name, such as ``"Ar"``, ``"Argon"``.
    mass, mass_unc : float
        Atomic weight in [amu] and the uncertainty.
    estimated_flag : str
        Generic field flagging estimated values.

    Attributes
    ----------
    A : int
    mass, mass_unc : float
    estimated_flag : str
    N : int
        A + Z

    Examples
    --------
    >>> isotope = Isotope(18, 38, "Ar", "Argon", 37.9627321040, 2.09e-07)
    >>> isotope.is_isotope
    True
    >>> isotope.A
    38
    >>> isotope.N
    20
    """

    is_isotope = True

    def __init__(
        self,
        atomic_number,
        mass_number,
        symbol,
        name,
        mass,
        mass_unc,
        estimated_flag="",
    ):
        super().__init__(symbol, name, atomic_number)
        self.A = mass_number
        self.mass = mass
        self.mass_unc = mass_unc
        self.estimated_flag = estimated_flag
        self.N = mass_number - atomic_number


def float_or_none(f):
    """Cast string ``f`` into a float, or None if it doesn't represent a number.

    Parameters
    ----------
    f : Any

    Returns
    -------
    float or NoneType
    """
    try:
        return float(f)
    except ValueError:
        return None


# list of all the element symbols recognised by pyvalem:
element_symbols = []
# pre-built mapping between element symbols and Atom instances:
atoms = {}
# Atom data is from Meija et al., "Atomic weights of the elements 2013
# (IUPAC Technical Report)", Pure Appl. Chem. 88(3), 265-291, 2016.
# See https://ciaaw.org/atomic-weights.htm
with pkg_resources.open_text("pyvalem", "_data_atomic_weights.txt") as fi:
    reader = csv.reader(fi, delimiter=",")
    header = ["Symbol", "Name", "Z", "atomic_weight", "atomic_weight_unc"]
    for row in reader:
        row = [val.strip() for val in row]
        if row == header:
            continue  # skip the header
        atom_dtypes = [str, str, int, float_or_none, float_or_none]
        atom_args = [dtype(val) for dtype, val in zip(atom_dtypes, row)]
        element_symbol = atom_args[0]
        element_symbols.append(element_symbol)
        atoms[element_symbol] = Atom(*atom_args)

# pre-built mapping between element isotopic symbols and Isotope instances:
isotopes = {}
# Isotope data is from the AME2016 Atomic Mass Evaluation reports,
# Huang et al., "The Ame2016 atomic mass evaluation (I)", Chinese Physics C41,
# 030002 (2017); Wang et al., "The Ame2016 atomic mass evaluation (II)",
# Chinese Physics C41, 030003 (2017).
# See http://amdc.impcas.ac.cn/masstables/Ame2016/mass16.txt
with pkg_resources.open_text("pyvalem", "_data_isotope_masses.txt") as fi:
    reader = csv.reader(fi, delimiter=",")
    header = ["Z", "A", "Symbol", "mass", "mass_unc", "estimated_flag"]
    iso_attribs = [
        "atomic_number",
        "mass_number",
        "symbol",
        "mass",
        "mass_unc",
        "estimated_flag",
    ]
    for row in reader:
        row = [val.strip() for val in row]
        if row == header:
            continue
        iso_dtypes = [int, int, str, float, float, str]
        iso_kwargs = {
            attr: dtype(val) for attr, dtype, val in zip(iso_attribs, iso_dtypes, row)
        }
        iso_name = "{}-{}".format(
            atoms[iso_kwargs["symbol"]].name, iso_kwargs["mass_number"]
        )
        iso_kwargs["name"] = iso_name
        iso_symbol = "{:d}{:s}".format(iso_kwargs["mass_number"], iso_kwargs["symbol"])
        iso_kwargs["symbol"] = iso_symbol
        isotopes[iso_symbol] = Isotope(**iso_kwargs)



"""
The VibrationalState class, representing a molecular vibrational state, with
methods for parsing a string into quantum numbers and labels, creating
an HTML representation of it, etc.
"""

import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError

integer = pp.Word(pp.nums)

# noinspection PyTypeChecker
vibrational_term = pp.Group(
    pp.Optional(integer.setResultsName("n"), default=1)
    + pp.Or(("v", "ν"))
    + integer.setResultsName("mode")
)
vibrational_config = (
    vibrational_term
    + pp.ZeroOrMore(pp.Suppress("+") + vibrational_term)
    + pp.StringEnd()
).leaveWhitespace()


class VibrationalTerm:
    def __init__(self, n, mode):
        self.n = int(n)
        self.mode = int(mode)

    def __repr__(self):
        return "{}ν{:d}".format(self.n if self.n != 1 else "", self.mode)

    @property
    def html(self):
        return "{}ν<sub>{:d}</sub>".format(self.n if self.n != 1 else "", self.mode)

    @property
    def latex(self):
        return "{}\\nu_{{{:d}}}".format(self.n if self.n != 1 else "", self.mode)

    def __lt__(self, other):
        return self.mode < other.mode


class VibrationalStateError(StateParseError):
    pass


class VibrationalState(State):
    def __init__(self, state_str):
        self.state_str = state_str.replace(" ", "")
        self.v = None
        self.polyatomic = None
        self.terms = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        if " + " in state_str:
            raise VibrationalStateError(
                'Spaces around "+" not allowed for'
                " polyatomic vibrational states: {}".format(state_str)
            )

        self.polyatomic = False

        def get_v(s_v):
            if s_v in ("*", "**", "***"):
                self.polyatomic = None
                return s_v
            return int(s_v)

        try:
            if self.state_str.startswith("v="):
                self.v = get_v(self.state_str[2:])
            else:
                self.v = get_v(self.state_str)
                self.state_str = "v={}".format(state_str)
        except ValueError:
            self.polyatomic = True
            try:
                parse_results = vibrational_config.parseString(self.state_str)
            except pp.ParseException:
                raise VibrationalStateError(
                    "Invalid vibrational state"
                    " configuration syntax: {0}".format(self.state_str)
                )

            self.terms = [
                VibrationalTerm(parsed_term["n"], parsed_term["mode"])
                for parsed_term in parse_results
            ]
            self.terms.sort()

    def __repr__(self):
        if self.v is not None:
            return "v={}".format(self.v)
        return "+".join([str(term) for term in self.terms])

    @property
    def html(self):
        if self.polyatomic:
            return " + ".join([term.html for term in self.terms])
        return str(self)

    @property
    def latex(self):
        if self.polyatomic:
            return " + ".join([term.latex for term in self.terms])
        return str(self)



"""Module with the `AtomicConfiguration` `State` subclass.

The `AtomicConfiguration` class, represents an atomic electronic configuration.
Methods are implemented for parsing a string into quantum numbers and
labels, creating an HTML representation of the term symbol, etc.
"""

import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError

integer = pp.Word(pp.nums).setParseAction(lambda t: int(t[0]))
nocc_integer = pp.Optional(pp.Word(pp.nums), default="1").setParseAction(
    lambda t: int(t[0])
)

# NB no "j" orbital.
atomic_orbital_symbols = tuple("spdfghiklmnoqrtuvwxyz")
noble_gases = ["He", "Ne", "Ar", "Kr", "Xe", "Rn"]
noble_gas_configs = {
    "He": "1s2",
    "Ne": "[He].2s2.2p6",
    "Ar": "[Ne].3s2.3p6",
    "Kr": "[Ar].3d10.4s2.4p6",
    "Xe": "[Kr].4d10.5s2.5p6",
    "Rn": "[Xe].4f14.5d10.6s2.6p6",
}
noble_gas_nelectrons = {"He": 2, "Ne": 10, "Ar": 18, "Kr": 36, "Xe": 54, "Rn": 86}

noble_gas = pp.oneOf(["[{}]".format(symbol) for symbol in noble_gases])

atom_orbital = pp.Group(
    integer.setResultsName("n")
    + pp.oneOf(atomic_orbital_symbols).setResultsName("lletter")
    + nocc_integer.setResultsName("nocc")
)

atom_config = (
    (atom_orbital | noble_gas)
    + pp.ZeroOrMore(pp.Suppress(".") + atom_orbital).leaveWhitespace()
    + pp.StringEnd()
).leaveWhitespace()


class AtomicOrbitalError(StateParseError):
    pass


class AtomicConfigurationError(StateParseError):
    pass


class AtomicOrbital:
    # noinspection PyUnresolvedReferences
    """A class representing a single atomic orbital.

    This is a building block of the `AtomicConfiguration` state class.

    Parameters
    ----------
    n : int
        The principal quantum number.
    l : int, optional
        The azimuthal quantum number (0, 1, 2, ..., n-1)
        At least one of {`l`, `lletter`} must be passed.
    nocc : int, default: 1
        Number of occupied electrons.
    lletter : str, optional
        The letter corresponding to l: 's', 'p', 'd', ... for l = 0, 1, 2, ...
        At least one of {`l`, `lletter`} must be passed.

    Attributes
    ----------
    n : int
    l : int
    lletter : str
    html
    latex

    Raises
    ------
    AtomicOrbitalError
        If inconsistent quantum numbers are passed to the constructor.
    """

    def __init__(self, n, l=None, nocc=0, lletter=None):
        self.n = n
        if l is None:
            self.lletter = lletter
            try:
                self.l = atomic_orbital_symbols.index(lletter)
            except IndexError:
                raise AtomicOrbitalError(
                    "Invalid azimuthal quantum"
                    ' number label, lletter="{}"'.format(lletter)
                )
        else:
            self.l = l
            try:
                self.lletter = atomic_orbital_symbols[l]
            except IndexError:
                raise AtomicOrbitalError(
                    "Invalid azimuthal quantum"
                    " number, l={}. Must be in range 0 < l < {}".format(
                        l, len(atomic_orbital_symbols)
                    )
                )
            if lletter and lletter != self.lletter:
                raise AtomicOrbitalError(
                    "Inconsistent azimuthal quantum"
                    " number, l={} and label, lletter={} specified".format(l, lletter)
                )

        self.nocc = nocc
        self._validate_atomic_orbital()

    def __repr__(self):
        if self.nocc != 1:
            return "{}{}{}".format(self.n, self.lletter, self.nocc)
        return "{}{}".format(self.n, self.lletter)

    @property
    def html(self):
        """Html representation of the AtomicOrbital instance.

        Returns
        -------
        str
        """
        if self.nocc != 1:
            return "{}{}<sup>{}</sup>".format(self.n, self.lletter, self.nocc)
        return "{}{}".format(self.n, self.lletter)

    @property
    def latex(self):
        """LaTeX representation of the AtomicOrbital instance.

        Returns
        -------
        str
        """
        if self.nocc != 1:
            return "{}{}^{{{}}}".format(self.n, self.lletter, self.nocc)
        return "{}{}".format(self.n, self.lletter)

    def _validate_atomic_orbital(self):
        """Validator method for the atomic orbital class.

        Checks if the quantum numbers passed are consistent and physical and
        raises `AtomicOrbitalError` if not.

        Raises
        ------
        AtomicOrbitalError
        """
        if self.l > self.n - 1:
            raise AtomicOrbitalError("l >= n in atomic orbital {}".format(self))
        if self.nocc < 0:
            raise AtomicOrbitalError(
                "Negative nocc = {} not allowed in"
                " orbital: {}".format(self.nocc, self.nocc)
            )
        if self.nocc > 2 * (2 * self.l + 1):
            raise AtomicOrbitalError(
                "Too many electrons in atomic" " orbital: {}".format(self)
            )


class AtomicConfiguration(State):
    # noinspection PyUnresolvedReferences
    """A class representing an atomic configuration.

    An atomic configuration is considered as a collection of occupied atomic
    orbitals: e.g. 1s2.2s2.2p6.3s1; this may be abbreviated by using [X] where
    [X] is a noble gas atom as a short cut for the closed-shell configuration
    of X.

    Parameters
    ----------
    state_str : str
        State string in the required format

    Attributes
    ----------
    state_str : str
    orbitals : list of `AtomicOrbital`
        This list does *not* include the closed-shell nobel gas configuration.
    noble_gas_config : str or NoneType
        For example "[Ne]"
    nelectrons : int
        Number of electrons in the whole configuration.
    html
    latex

    Raises
    ------
    AtomicConfigurationError
        If the state cannot be parsed from the `state_str`.

    Notes
    -----
    The ``__repr__`` method is overloaded to provide a *canonicalised* representation
    of the state.
    The idea is that two states representing the same physical entity will have the
    same ``repr(state)`` representation.

    Examples
    --------
    >>> ac1 = AtomicConfiguration("1s2.2s2.2p6.3p1")
    >>> outer_orbital = ac1.orbitals[-1]
    >>> type(outer_orbital)
    <class 'pyvalem.states.atomic_configuration.AtomicOrbital'>

    >>> outer_orbital.n, outer_orbital.l, outer_orbital.nocc
    (3, 1, 1)

    >>> repr(ac1)
    '1s2.2s2.2p6.3p'

    >>> ac2 = AtomicConfiguration("[Ne].3p")
    >>> repr(ac2)
    '1s2.2s2.2p6.3p'
    """

    def __init__(self, state_str):
        self.state_str = state_str
        self.orbitals = []
        self.noble_gas_config = None
        self.nelectrons = 0
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        """Parses the `AtomicConfiguration` instance from the supplied `state_str`.

        Parameters
        ----------
        state_str : str

        Raises
        ------
        AtomicConfigurationError
            If the `state_str` cannot be parsed into a valid state.
        """
        try:
            parse_results = atom_config.parseString(state_str)
        except pp.ParseException:
            raise AtomicConfigurationError(
                "Invalid atomic electronic"
                " configuration syntax: {0}".format(state_str)
            )

        # Expand out noble gas notation, if used, and check that the
        # subshells 1s, 2s, 2p, ... are unique.
        subshells = self._expand_noble_gas_config(state_str)
        subshells = [subshell[:2] for subshell in subshells.split(".")]

        for i, parsed_orbital in enumerate(parse_results):
            if not i and type(parsed_orbital) == str:
                # Noble-gas notation for first atomic orbital
                self.noble_gas_config = parsed_orbital
                # NB strip '[' and ']' from parsed_orbital of the form '[Ar]'.
                self.nelectrons = noble_gas_nelectrons[parsed_orbital[1:-1]]
                continue
            # Create a validated AtomicOrbital object for this orbital.
            try:
                orbital = AtomicOrbital(
                    n=parsed_orbital["n"],
                    lletter=parsed_orbital["lletter"],
                    nocc=parsed_orbital["nocc"],
                )
                self.nelectrons += orbital.nocc
            except AtomicOrbitalError as err:
                raise AtomicConfigurationError(err)
            self.orbitals.append(orbital)

        # Check that the subshells specified are unique
        if len(subshells) != len(set(subshells)):
            raise AtomicConfigurationError("Repeated subshell in {0}".format(state_str))

    @property
    def html(self):
        """See the `State` base class."""
        html_chunks = []
        if self.noble_gas_config:
            html_chunks.append(self.noble_gas_config)
        for orbital in self.orbitals:
            html_chunks.append(orbital.html)
        return "".join(html_chunks)

    @property
    def latex(self):
        """See the `State` base class."""
        latex_chunks = []
        if self.noble_gas_config:
            latex_chunks.append(r"\mathrm{{{}}}".format(self.noble_gas_config))
        for orbital in self.orbitals:
            latex_chunks.append(orbital.latex)
        return "".join(latex_chunks)

    def _expand_noble_gas_config(self, config):
        """Recursively expand out the noble gas notation to orbitals.

        For example:
        '[He].2s1' -> '1s2.2s2',
        '[Xe].4f7' -> '1s2.2s2.2p6.3s2.3p6.3d10.4s2.4p6.4d10.5s2.5p6.4f7'

        Parameters
        ----------
        config : str
            Closed shell nobel gas configuration, in the form of '[Ne]', '[Xe]', etc.

        Returns
        -------
        str
        """

        if config[0] != "[":
            return config
        return (
            self._expand_noble_gas_config(noble_gas_configs[config[1:3]]) + config[4:]
        )

    def __repr__(self):
        """See the `State` base class."""
        if self.noble_gas_config:
            state_repr = self._expand_noble_gas_config(self.noble_gas_config)
            if self.orbitals:
                state_repr += "." + ".".join(repr(orbital) for orbital in self.orbitals)
            return state_repr
        else:
            return ".".join(repr(orbital) for orbital in self.orbitals)



"""
The AtomicTermSymbol class, representing an atomic term symbol, with
methods for parsing a string into quantum numbers and labels, creating
an HTML representation of the term symbol, etc.
"""

import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError
from pyvalem._utils import parse_fraction, float_to_fraction

atom_L_symbols = "S P D F G H I K L M N O Q R T U V W X Y Z".split()

integer = pp.Word(pp.nums)

atom_Smult = integer.setResultsName("Smult")
atom_Lletter = pp.oneOf(atom_L_symbols).setResultsName("Lletter")
atom_Jstr = (
    integer + pp.Optional(pp.Suppress("/") + "2") + pp.StringEnd()
).setResultsName("Jstr")
atom_parity = pp.Literal("o").setResultsName("parity")
atom_term = (
    atom_Smult
    + atom_Lletter
    + pp.Optional(atom_parity)
    + pp.Optional(pp.Suppress("_") + atom_Jstr)
    + pp.StringEnd()
)


class AtomicTermSymbolError(StateParseError):
    pass


class AtomicTermSymbol(State):
    def __init__(self, state_str):
        self.state_str = state_str
        self.Smult = None
        self.S = None
        self.Lletter = None
        self.L = None
        self.parity = None
        self.J = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        try:
            components = atom_term.parseString(state_str)
        except pp.ParseException:
            raise AtomicTermSymbolError(
                "Invalid atomic term symbol syntax:" " {0}".format(state_str)
            )
        self.Smult = int(components.Smult)
        self.S = (self.Smult - 1) / 2.0
        self.Lletter = components.Lletter
        self.L = atom_L_symbols.index(components.Lletter)
        self.parity = components.get("parity")
        try:
            self.J = parse_fraction(components.Jstr)
        except ValueError as err:
            raise AtomicTermSymbolError(err)
        if self.J is not None:
            self._validate_j()

    def _validate_j(self):
        s_is_half_integer = int(2 * self.S) % 2
        j_is_half_integer = int(2 * self.J) % 2
        if s_is_half_integer != j_is_half_integer:
            raise AtomicTermSymbolError(
                "J={} is invalid for S={}.".format(self.J, self.S)
            )
        if not abs(self.L - self.S) <= self.J <= self.L + self.S:
            raise AtomicTermSymbolError(
                "Invalid atomic term symbol: {0:s}"
                " |L-S| <= J <= L+S must be satisfied.".format(self.state_str)
            )

    @property
    def html(self):
        html_chunks = ["<sup>{0:d}</sup>{1:s}".format(self.Smult, self.Lletter)]
        if self.parity:
            html_chunks.append("<sup>o</sup>")
        if self.J is not None:
            j_str = float_to_fraction(self.J)
            html_chunks.append("<sub>{0:s}</sub>".format(j_str))
        return "".join(html_chunks)

    @property
    def latex(self):
        latex_chunks = ["{{}}^{{{}}}\mathrm{{{}}}".format(self.Smult, self.Lletter)]
        if self.parity:
            latex_chunks.append("^o")
        if self.J is not None:
            j_str = float_to_fraction(self.J)
            latex_chunks.append("_{{{}}}".format(j_str))
        return "".join(latex_chunks)

    def __repr__(self):
        return self.state_str



"""
The KeyValuePair class, representing an arbitrary atomic or molecular state,
with methods for parsing it from a string and outputting its HTML
representation, etc.
"""
from pyvalem.states._base_state import State, StateParseError


class KeyValuePairError(StateParseError):
    pass


class KeyValuePair(State):
    """
    A class representing an state as an arbitrary key/value pair.

    Any atomic or molecular state that doesn't fit into any of the other
    explicit categories available in PyValem can be represented as a
    key/value pair, parsed in the form 'key=value'.
    """

    multiple_allowed = True

    def __init__(self, state_str):
        self.state_str = None
        self.key = None
        self.value = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):
        """
        Parse state_str into a KeyValuePair object.

        Whitespace is not tolerated in the input, i.e only 'key=value' and not
        'key = value' are allowed. No spaces are inserted in the output.
        """

        if any(c.isspace() for c in state_str):
            raise KeyValuePairError(
                "No whitespace allowed in key-value pair: {}".format(state_str)
            )

        try:
            key, value = state_str.split("=")
        except ValueError:
            raise KeyValuePairError("Invalid key-value pair: {}.".format(state_str))
        self.key = key.strip()
        self.value = value.strip()

        self.state_str = "{}={}".format(self.key, self.value)

    def __repr__(self):
        return self.state_str



"""
The StatefulSpecies class representing a chemical species (e.g. atom,
molecule), optionally with with one or more State objects associated with it.

The Formula of the StatefulSpecies is separated from its States by whitespace;
States are separated from each other by semicolons (;) or whitespace.
"""
from pyvalem.states.atomic_configuration import AtomicConfiguration
from pyvalem.states.diatomic_molecular_configuration import (
    DiatomicMolecularConfiguration,
)
from .formula import Formula
from pyvalem.states.key_value_pair import KeyValuePair
from pyvalem.states._state_parser import state_parser, STATES


class StatefulSpeciesError(Exception):
    pass


class StatefulSpecies:
    def __init__(self, s):
        s = s.strip()
        if " " not in s:
            # No states, just a Formula
            self.states = []
            self.formula = Formula(s)
            return

        i = s.index(" ")
        self.formula = Formula(s[:i])
        s = s.replace(";", " ").replace(",", " ")
        self.states = state_parser(s[i + 1 :].split())

        self._verify_states()

    def __repr__(self):
        """Return a canonical text representation of the StatefulSpecies."""
        if self.states:
            states_sorted = sorted(
                self.states, key=lambda state: (STATES[type(state)], repr(state))
            )
            return "{} {}".format(
                self.formula, ";".join(repr(state) for state in states_sorted)
            )
        return self.formula.__repr__()

    def __eq__(self, other):
        """
        Two StatefulSpecies are equal if they have the same Formula and the
        equal States (in any order).
        """

        return self.formula == other.formula and set(self.states) == set(other.states)

    def __hash__(self):
        return hash(repr(self))

    def _verify_states(self):
        """Check that multiple states are not given where this is not allowed.

        For example, a StatefulSpecies may not have two atomic configurations.

        """

        state_class_counts = {}
        for state in self.states:
            state_class_counts.setdefault(state.__class__, 0)
            state_class_counts[state.__class__] += 1

        for state_class in state_class_counts:
            if not state_class.multiple_allowed and state_class_counts[state_class] > 1:
                raise StatefulSpeciesError(
                    "Multiple states of type {}"
                    " specified for {}".format(state_class.__name__, self)
                )
        return True

    def get_states_of_type(self, state_class):
        return [s for s in self.states if s.__class__ == state_class]

    def verify_atomic_configuration(self):
        atomic_configurations = self.get_states_of_type(AtomicConfiguration)
        if len(atomic_configurations) != 1:
            raise StatefulSpeciesError(
                "Multiple AtomicConfigurations" " specified for {}".format(self)
            )
        if not atomic_configurations:
            return True
        atomic_configuration = atomic_configurations[0]
        if self.formula.natoms != 1:
            return True

        atom = next(iter(self.formula.atoms))
        nelectrons = atom.Z - self.formula.charge
        if nelectrons != atomic_configuration.nelectrons:
            raise StatefulSpeciesError(
                "Incorrect number of electrons for"
                " {} in atomic configuration {}".format(
                    self.formula, atomic_configuration
                )
            )
        return True

    def verify_diatomic_inversion_parity(self):
        if self.formula.natoms != 2:
            return True
        homonuclear_diatomic = len(self.formula.atoms) == 1
        diatomic_configurations = self.get_states_of_type(
            DiatomicMolecularConfiguration
        )
        if not diatomic_configurations:
            return True
        if len(diatomic_configurations) != 1:
            raise StatefulSpeciesError(
                "Multiple DiatomicConfigurations" " specified for {}".format(self)
            )
        diatomic_configuration = diatomic_configurations[0]
        has_parity_label = all(
            orbital.symbol[-1] in "gu" for orbital in diatomic_configuration.orbitals
        )
        if has_parity_label == homonuclear_diatomic:
            return True
        s_homohetero = "homo" if homonuclear_diatomic else "hetero"
        raise StatefulSpeciesError(
            "Incorrect use of inversion parity label"
            " (u/g) for {}nuclear diatomic {}".format(s_homohetero, self.formula)
        )

    def verify_multiple_key_value_pairs(self):
        key_value_pairs = self.get_states_of_type(KeyValuePair)
        keys = [key_value_pair.key for key_value_pair in key_value_pairs]
        if len(set(keys)) != len(keys):
            raise StatefulSpeciesError(
                "Multiple key-value pairs with the same"
                " key specified for {}".format(self)
            )

    @property
    def html(self):
        if not self.states:
            return self.formula.html
        return "{} {}".format(self.formula.html, " ".join(s.html for s in self.states))

    @property
    def latex(self):
        if not self.states:
            return self.formula.latex
        return r"{} \; {}".format(
            self.formula.latex, r" \; ".join(s.latex for s in self.states)
        )



from setuptools import setup, find_packages
from pathlib import Path

root = Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (root / "README.rst").read_text(encoding="utf-8")

setup(
    name="pyvalem",
    version="2.5.6",
    description="A package for managing simple chemical species and states",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/xnx/pyvalem",
    author="Christian Hill",
    author_email="ch.hill@iaea.org",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ],
    keywords="chemistry, formula, species, state, reaction",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        "pyparsing>=2.3",
        'importlib-resources>=1.0; python_version < "3.7.0"',
    ],
    extras_require={"dev": ["black", "pytest-cov", "tox", "ipython"]},
    # package_data will include all the resolved globs into both the wheel and sdist
    package_data={"pyvalem": ["*.txt"]},
    # no need for MANIFEST.in, which should be reserved only for build-time files
    project_urls={
        "Bug Reports": "https://github.com/xnx/pyvalem/issues",
    },
)



"""
This module defines the `Formula` class, representing a chemical formula (of an atom,
isotope, ion, molecule, molecular-ion, isotopologue, etc.) and associated
exceptions.

Quantum states are not defined in the `Formula` framework, for chemical species
with quantum states, look at the `StatefulSpecies` class.

Examples
--------
>>> from pyvalem.formula import Formula
>>> Formula("H2O+").charge
1

>>> dict(Formula("CH4").atom_stoich.items())
{'C': 1, 'H': 4}
"""

import re
from collections import defaultdict

import pyparsing as pp

from ._special_cases import special_cases
from .atom_data import element_symbols, atoms, isotopes

element = pp.oneOf(element_symbols)
# TODO: don't allow leading 0
integer = pp.Word(pp.nums)
integer_or_x = integer | pp.Literal("x")
plusminus = pp.Literal("+") | pp.Literal("-")

# An isotope looks like '(1H)', '(13C)', etc.; strip the parentheses.
isotope = (
    pp.Group(
        pp.Suppress(pp.Literal("(")) + integer + element + pp.Suppress(pp.Literal(")"))
    )
    | pp.Literal("D")
    | pp.Literal("T")
)

# some named components of a formula
# stoich comes before or after a "bare" element symbol, e.g. 3Br2.
pp_stoich = pp.Optional(integer_or_x, default="1").setResultsName("stoich")
# prestoich comes before a bracketed element group, e.g. 2(NH3).
pp_prestoich = pp.Optional(integer, default="1").setResultsName("prestoich")
# poststoich comes after a bracketed element group, e.g. (H2O)3.
pp_poststoich = pp.Optional(integer_or_x, default="1").setResultsName("poststoich")
# the charge is always defined and defaults to '0'.
pp_charge = pp.Optional(
    pp.Combine(pp.Group(plusminus + pp.Optional(integer, default="1"))), default="0"
).setResultsName("charge")
# An elementRef is either an element symbol or an isotope symbol plus a
# stoichiometry which is 1 if not given.
# NB check for element symbol first to catch Dy, Ti, etc. before parsing as
# hydrogen isotopes D or T.
elementRef = pp.Group((element | isotope).setResultsName("atom_symbol") + pp_stoich)
# A chemicalFormula is a series of elementRefs.
chemicalFormula = pp.Group(pp.OneOrMore(elementRef)).setResultsName("atoms")
# Radicals must be specified with the unicode character '·' (U+00B7).
radicalDot = pp.Optional("·").setResultsName("radical")

# A chargedChemicalFormula is a chemicalFormula with an optional pre-
# stoichiometry, radical dot and charge.
chargedChemicalFormula = pp.Group(pp_stoich + chemicalFormula + radicalDot + pp_charge)

# A bracketedChemicalFormula is essentially a chargedChemicalFormula inside
# parentheses with optional pre- and post-stoichiometries.
left_bracket = pp.Literal("(")
right_bracket = pp.Literal(")")
bracketedChemicalFormula = pp.Group(
    pp_prestoich
    + pp.Suppress(left_bracket)
    + chemicalFormula
    + radicalDot
    + pp_charge
    + pp.Suppress(right_bracket)
    + pp_poststoich
)

# A chemical_formula_fragment is either one of the above ChemicalFormula
# types, optionally prefixed with a period ('.')
chemical_formula_fragment = pp.Optional(pp.Suppress(".")) + (
    chargedChemicalFormula | bracketedChemicalFormula
)
chemical_formula_fragments = pp.OneOrMore(chemical_formula_fragment).setResultsName(
    "formula"
)

# These are the formula prefix tokens we recognise and their slugs
# NB be careful to ensure that the slugs remain case-insensitive!
# For example, '(S)-' -> 'S-' but 's-' -> 'syn-'
prefix_tokens = {
    "(+)": "p",
    "(-)": "m",
    "(±)": "pm",
    "D": "D",
    "L": "L",
    "(R)": "R",
    "(S)": "S",
    "(E)": "E",
    "(Z)": "Z",
    "c": "c",
    "l": "l",
    "cis": "cis",
    "trans": "trans",
    "s": "syn",
    "a": "anti",
    "Δ": "Delta",
    "Λ": "Lambda",
    "α": "alpha",
    "β": "beta",
    "γ": "gamma",
    "n": "n",
    "i": "i",
    "t": "t",
    "neo": "neo",
    "sec": "sec",
    "o": "o",
    "m": "m",
    "p": "p",
    "ortho": "ortho",
    "meta": "meta",
    "para": "para",
}
# The LaTeX equivalents of these prefix tokens.
latex_prefix_dict = {
    "(±)": r"\pm",
    "Δ": r"\Delta",
    "Λ": r"\Lambda",
    "α": r"\alpha",
    "β": r"\beta",
    "γ": r"\gamma",
}

# also allow a comma-separated list of integers, e.g. 1,1,2-
prefix_parser = pp.delimitedList(pp.OneOrMore(integer), combine=True)
keys_list = list(prefix_tokens.keys())
for pt in keys_list:
    prefix_parser |= pp.Keyword(pt)
# Any number of prefix tokens may appear, separated by a hyphen
prefix_list_parser = (
    pp.delimitedList(prefix_parser, delim="-") + pp.Suppress(pp.Literal("-"))
).setResultsName("prefix")

# Finally, a complexChemicalFormula is a prefix, followed by one or more
# chemical_formula_fragments.
complexChemicalFormula = (
    pp.Optional(prefix_list_parser)
    + chemical_formula_fragments
    + pp_charge
    + pp.StringEnd()
)
# replace '+' and '-' with 'p' and 'm' to make slugs for e.g. URLs.
slug_charge_sign = {"+": "p", "-": "m"}


class FormulaError(Exception):
    pass


class FormulaParseError(FormulaError):
    pass


class Formula:
    """A class representing a chemical formula without states.

    Methods are implemented for the `Formula` from a compatible string and providing
    *html* and *LaTeX* representations, atomic stoichiometry, species mass, number of
    atoms, or html-save slug.

    Parameters
    ----------
    formula : str
        PyValem-compatible string. No ``"_"`` or ``"^"`` symbols to indicate subscripts
        and superscripts. Brackets are allowed, as well as several common prefixes.
        Charges are provided by ``"+"/"-"``, or ``"+n"/"-n"``, where `n` is the charge

    Attributes
    ----------
    formula : str
        The formula string as passed to the constructor.
    atoms : set of `Atom`
    atom_stoich : dict[str, int]
    charge : int
        Charge in [e].
    natoms : int
        Number of atoms of the `Formula`.
    rmm, mass : float
        Both are the mass in [amu] (one is alias for the other).
    html, latex, slug : str
        Different string representations of the `Formula`. The ``slug`` is a url-safe
        representation.

    Raises
    ------
    FormulaParseError
        When an incompatible `formula` string is passed into the constructor

    Notes
    -----
    The ``__repr__`` method is overloaded to provide a *canonicalised* representation
    of the formula. The idea is that two formulas representing the same physical entity
    will have the same ``repr(formula)`` representation.

    Examples
    --------
    >>> Formula("H2+")  # a simple chemical formula instantiation
    H2+

    >>> Formula("(1H)2(16O)")  # isotopologue instantiation
    (1H)2(16O)

    >>> Formula("ortho-C6H4(CH3)2"), Formula("W+42")  # more complicated situations
    (ortho-C6H4(CH3)2, W+42)

    >>> # Attributes available:
    >>> dict(Formula("CH4").atom_stoich)
    {'C': 1, 'H': 4}
    >>> Formula("(2H)(3H)+").html
    '<sup>2</sup>H<sup>3</sup>H<sup>+</sup>'
    >>> Formula("Ar").mass
    39.95

    >>> # Incompatible formula string
    >>> Formula("Argon")
    Traceback (most recent call last):
    ...
    pyvalem.formula.FormulaParseError: Invalid formula syntax: Argon
    """

    def __init__(self, formula):
        """Initialize the Formula object by parsing the string argument formula."""
        self.formula = formula
        self.atoms = set()
        self.atom_stoich = defaultdict(int)
        self.charge = 0
        self.natoms = 0
        self.rmm = 0.0
        self.html = ""
        self.latex = ""
        self.slug = ""
        self.mass = 0.0
        self._parse_formula(formula)

    @staticmethod
    def _make_prefix_html(prefix_list):
        """Make the prefix HTML: D- and L- prefixes get written in small caps"""
        prefix = "-".join(prefix_list)
        prefix = prefix.replace("D", '<span style="font-size: 80%;">D</span>')
        prefix = prefix.replace("L", '<span style="font-size: 80%;">L</span>')
        return "{}-".format(prefix)

    @staticmethod
    def _make_prefix_latex(prefix_list):
        """Make the prefix LaTeX."""
        latex_prefixes = [latex_prefix_dict.get(pr, pr) for pr in prefix_list]
        prefix = "-".join(latex_prefixes)
        return "{}-".format(prefix)

    @staticmethod
    def _make_prefix_slug(prefix_list):
        """Make the prefix slug

        Commas are replaced by underscores and non-ASCII
        characters swapped out according to the entries in the prefix_tokens
        dictionary. For example,
            1,1,3- -> 1_1_3-
            (±)- -> pm-
            (α)- -> alpha-
        """
        slug_prefix_tokens = []
        for prefix_token in prefix_list:
            # TODO don't allow matches including numbers with leading zeros
            if re.match(r"^\d+(,\d+)*$", prefix_token):
                # this prefix token is a comma-separated list of numbers
                slug_prefix_token = prefix_token.replace(",", "_")
            else:
                # select the slug version of this prefix from prefix_tokens
                try:
                    slug_prefix_token = prefix_tokens[prefix_token]
                except KeyError:
                    raise FormulaParseError(
                        "Unrecognised formula prefix" " token: {}".format(prefix_token)
                    )
            slug_prefix_tokens.append(slug_prefix_token)
        slug_prefix = "_".join(slug_prefix_tokens)
        return "{}__".format(slug_prefix)

    def _parse_formula(self, formula):
        """Parse the formula string into a Formula object.

        The main method of the class, populates all the instance attributes.

        Parameters
        ----------
        formula : str
            See the class docstring.

        Raises
        ------
        FormulaParseError
        """
        if formula == "D-":
            # This is a not-ideal way to deal with the fact that D- breaks
            # the parser due to a clash with the D- prefix.
            self._parse_formula("D-1")
            return

        if any(s in formula for s in ("++", "--", "+-", "-+")):
            raise FormulaParseError("Invalid formula syntax: {}".format(formula))

        # We make a particular exception for various special cases, including
        # photons, electrons, positrons and "M", denoting an unspecified
        # "third-body" in many reactions. Note that M does not have a defined
        # charge or mass.
        if formula == "e":
            # Quietly convert e to e-.
            self.formula = formula = "e-"
        if formula in special_cases:
            for attr, val in special_cases[formula].items():
                setattr(self, attr, val)
            return

        try:
            moieties = complexChemicalFormula.parseString(formula)
        except pp.ParseException:
            raise FormulaParseError("Invalid formula syntax: %s" % formula)

        html_chunks = []
        latex_chunks = []
        slug_chunks = []
        # calculate relative molecular mass as the sum of the atomic weights

        # make the prefix html and slug
        if "prefix" in moieties.keys():
            html_chunks.append(self._make_prefix_html(moieties["prefix"]))
            latex_chunks.append(self._make_prefix_latex(moieties["prefix"]))
            slug_chunks.append(self._make_prefix_slug(moieties["prefix"]))

        moieties, total_charge = moieties["formula"], int(moieties["charge"])
        nmoieties = len(moieties)
        for i, moiety in enumerate(moieties):
            poststoich = 0
            if "prestoich" in moiety.keys():
                # bracketed fragment, e.g. (OH)2, 3(HO2+), ...
                prestoich = int(moiety["prestoich"])
                if prestoich > 1:
                    slug_chunks.append(moiety["prestoich"])
                    html_chunks.append(moiety["prestoich"])
                    latex_chunks.append(moiety["prestoich"])
                html_chunks.append("(")
                latex_chunks.append("(")
                slug_chunks.append("_l_")
                poststoich = int(moiety["poststoich"])
                stoich = prestoich * poststoich
            else:
                # unbracketed fragment, e.g. H2O, 2NH3, ...
                stoich = int(moiety["stoich"])
                if stoich > 1:
                    slug_chunks.append(moiety["stoich"])
                    html_chunks.append(moiety["stoich"])
                    latex_chunks.append(moiety["stoich"])
            charge = int(moiety["charge"])
            self.charge += charge * stoich
            for atom in moiety["atoms"]:
                atom_symbol, atom_stoich = atom

                if atom_symbol == "D":
                    atom_symbol = isotope.parseString("(2H)")[0]
                if atom_symbol == "T":
                    atom_symbol = isotope.parseString("(3H)")[0]
                if isinstance(atom_symbol, pp.ParseResults):
                    # we got an isotope in the form '(zSy)' with z the mass
                    # number so symbol is the ParseResults ['z', 'Sy']:
                    mass_number, atom_symbol = (int(atom_symbol[0]), atom_symbol[1])
                    symbol_html = "<sup>%d</sup>%s" % (mass_number, atom_symbol)
                    symbol_latex = r"{{}}^{{{0:d}}}\mathrm{{{1:s}}}".format(
                        mass_number, atom_symbol
                    )
                    atom_symbol = "%d%s" % (mass_number, atom_symbol)
                    slug_chunks.append("-%s" % atom_symbol)
                else:
                    mass_number = None
                    symbol_html = atom_symbol
                    symbol_latex = r"\mathrm{{{0:s}}}".format(atom_symbol)
                    slug_chunks.append(atom_symbol)

                if mass_number is not None:
                    # We have an isotope, specified as '(zSy)'.
                    try:
                        atom = isotopes[atom_symbol]
                    except KeyError:
                        raise FormulaParseError(
                            "Unknown isotope symbol"
                            " {} in formula {}".format(atom_symbol, formula)
                        )
                else:
                    # A regular element symbol for an atom.
                    try:
                        atom = atoms[atom_symbol]
                    except KeyError:
                        raise FormulaParseError(
                            "Unknown element symbol"
                            " {} in formula {}".format(atom_symbol, formula)
                        )

                self.atoms.add(atom)

                try:
                    i_atom_stoich = int(atom_stoich)
                    total_atom_stoich = i_atom_stoich * stoich
                    self.natoms += total_atom_stoich
                    try:
                        self.rmm += atom.mass * total_atom_stoich
                    except TypeError:
                        # No atomic weight for some elements (e.g. Tc, Pm)
                        self.rmm = None
                except ValueError:
                    # Some formulas don't have well-defined stoichiometries,
                    # e.g. 'CFx'
                    total_atom_stoich = None
                    self.natoms = self.rmm = None
                try:
                    self.atom_stoich[atom_symbol] += total_atom_stoich
                except TypeError:
                    # There's not much we can do for the stoichiometry if this
                    # atom occurs with an undefined value.
                    # noinspection PyTypeChecker
                    self.atom_stoich[atom_symbol] = None

                html_chunks.append(symbol_html)
                latex_chunks.append(symbol_latex)
                if atom_stoich != "1":
                    html_chunks.append("<sub>{}</sub>".format(atom_stoich))
                    latex_chunks.append("_{{{0}}}".format(atom_stoich))
                    slug_chunks.append(str(atom_stoich))

            (
                moiety_charge_html,
                moiety_charge_latex,
                moiety_charge_slug,
            ) = self._get_charge_reps(charge)
            html_chunks.append(moiety_charge_html)
            latex_chunks.append(moiety_charge_latex)
            slug_chunks.append(moiety_charge_slug)
            # if i < nmoieties-1 and nmoieties != 1:
            #    slug_chunks.append('_d_')

            if "poststoich" in moiety.keys():
                html_chunks.append(")")
                latex_chunks.append(")")
                slug_chunks.append("_r_")
                # if i == nmoieties and nmoieties != 1:
                #    slug_chunks.append('_d_')
                if poststoich > 1:
                    html_chunks.append("<sub>{:d}</sub>".format(poststoich))
                    latex_chunks.append("_{{{0:d}}}".format(poststoich))
                    slug_chunks.append("{:d}".format(poststoich))

            if "radical" in moiety.keys():
                html_chunks.append("&#183;")
                latex_chunks.append(r"\cdot")
                slug_chunks.append("_dot")
            if i != nmoieties - 1:
                slug_chunks.append("-")

        if total_charge:
            # TODO Refactor and test further
            self.charge = int(total_charge)
            (
                moiety_charge_html,
                moiety_charge_latex,
                moiety_charge_slug,
            ) = self._get_charge_reps(self.charge)
            html_chunks.append(moiety_charge_html)
            latex_chunks.append(moiety_charge_latex)
            slug_chunks.append(moiety_charge_slug)

        self.html = "".join(html_chunks)
        self.latex = "".join(latex_chunks)
        # strip the leading '-' if the formula began with an isotope
        self.slug = "".join(slug_chunks).lstrip("-")

        self.mass = self.rmm

    @staticmethod
    def _get_charge_reps(charge):
        if charge:
            s_charge = ""
            if abs(charge) > 1:
                s_charge = str(abs(charge))
            s_charge_sign = "+"
            if charge < 0:
                s_charge_sign = "-"
            moiety_charge_html = "<sup>{:s}{:s}</sup>".format(s_charge, s_charge_sign)
            moiety_charge_latex = "^{{{0:s}{1:s}}}".format(s_charge, s_charge_sign)
            moiety_charge_slug = "_{:s}{:s}".format(
                slug_charge_sign[s_charge_sign], s_charge
            )
            return moiety_charge_html, moiety_charge_latex, moiety_charge_slug
        return "", "", ""

    def __repr__(self):
        if self.formula == "hv":
            return "hν"
        return self.formula

    def __eq__(self, other):
        return repr(self.formula) == repr(other.formula)

    def _stoichiometric_formula_atomic_number(self):
        """Return a list of atoms/isotopes and their stoichiometries.

        The returned list is sorted in order of increasing atomic number.
        """

        atom_strs = []
        for atom in sorted(self.atoms, key=lambda e: (e.Z, e.mass)):
            if atom.is_isotope:
                symbol = "({})".format(atom.symbol)
            else:
                symbol = atom.symbol
            atom_strs.append(
                self._get_symbol_stoich(symbol, self.atom_stoich[atom.symbol])
            )
        return atom_strs

    def _stoichiometric_formula_alphabetical(self):
        """Return a list of atoms/isotopes and their stoichiometries.

        The returned list is sorted in alphabetical order.
        """
        atom_strs = self._stoichiometric_formula_atomic_number()
        atom_strs.sort()
        return atom_strs

    def _stoichiometric_formula_hill(self):
        """Return a list of atoms/isotopes and their stoichiometries.

        The returned list is sorted in "Hill notation": first carbon, then
        hydrogen, then all other chemical elements in alphabetical order.
        If the species contains H but no C, all elements are listed in
        alphabetical order, including hydrogen.
        """
        c_h_strs = []
        atom_strs = []
        contains_c = False
        for atom in sorted(self.atoms, key=lambda e: (e.Z, e.mass)):
            if atom.is_isotope:
                symbol = "({})".format(atom.symbol)
            else:
                symbol = atom.symbol
            if symbol == "C":
                c_h_strs.insert(0, self._get_symbol_stoich("C", self.atom_stoich["C"]))
                contains_c = True
            elif symbol == "H":
                c_h_strs.insert(1, self._get_symbol_stoich("H", self.atom_stoich["H"]))
            else:
                atom_strs.append(
                    self._get_symbol_stoich(symbol, self.atom_stoich[atom.symbol])
                )

        if not contains_c:
            atom_strs = c_h_strs + atom_strs
            c_h_strs = []
        atom_strs.sort()
        return c_h_strs + atom_strs

    def stoichiometric_formula(self, fmt="atomic number"):
        """Return a string representation of the stoichiometric formula.

        The formula is given in one of the formats specified by the fmt argument:
        "atomic number": order atoms by increasing atomic number
        "alphabetical" : order atoms in alphabetical order of atomic symbol
        "hill": first C, then H, then all other chemical elements in alphabetical order,
                if C not present, all alphabetical.

        Parameters
        ----------
        fmt : {"atomic number", "alphabetical", "hill"}

        Returns
        -------
        str
            Stoichiometric formula (with charges)

        Examples
        --------
        >>> Formula("L-CH3CH(NH2)CO2H").stoichiometric_formula("atomic number")
        'H7C3NO2'

        >>> Formula("L-CH3CH(NH2)CO2H+").stoichiometric_formula("alphabetical")
        'C3H7NO2+'
        """
        # Special cases
        if self.formula in {"M", "e-", "e+"}:
            return self.formula
        if self.formula in {"hv", "hν"}:
            # Special case for the photon
            return "hν"

        fmt = fmt.lower()
        if fmt == "atomic number":
            atom_strs = self._stoichiometric_formula_atomic_number()
        elif fmt == "alphabetical":
            atom_strs = self._stoichiometric_formula_alphabetical()
        elif fmt == "hill":
            atom_strs = self._stoichiometric_formula_hill()
        else:
            raise FormulaError(
                "Unsupported format for stoichiometric formula output: {}".format(fmt)
            )

        # finally, add on the charge string, e.g. '', '-', '+2', ...
        atom_strs.append(self._get_charge_string())
        return "".join(atom_strs)

    @staticmethod
    def _get_symbol_stoich(symbol, stoich):
        """Return Xn for element symbol X and stoichiometry n, unless n is 1,
        in which case, just return X.
        """
        if stoich is None:
            return "{}x".format(symbol)
        if stoich > 1:
            return "{}{:d}".format(symbol, stoich)
        return symbol

    def _get_charge_string(self):
        """Return the string representation of the charge: '+', '-', '+2', '-3', etc."""
        if not self.charge:
            return ""
        if self.charge > 0:
            if self.charge > 1:
                return "+{:d}".format(self.charge)
            return "+"
        if self.charge == -1:
            return "-"
        return str(self.charge)



"""
The RotationalState class, representing a rotational quantum number, with
methods for parsing a string into a value and an HTML representation, etc.
"""

import pyparsing as pp

from pyvalem.states._base_state import State, StateParseError

integer = pp.Word(pp.nums)
integer_string = (integer + pp.StringEnd()).setResultsName("integer")
frac_string = (
    integer + pp.Suppress("/") + pp.Suppress(pp.Literal("2")) + pp.StringEnd()
).setResultsName("fraction_half")
decimal_string = (
    integer + pp.Suppress(".") + pp.Suppress(pp.Literal("5")) + pp.StringEnd()
).setResultsName("decimal_half")

Jstr = integer_string | frac_string | decimal_string


class RotationalStateError(StateParseError):
    pass


class RotationalState(State):
    def __init__(self, state_str):
        self.state_str = None
        self.J = None
        self._parse_state(state_str)

    def _parse_state(self, state_str):

        try:
            k, v = state_str.split("=")
            if k.strip() != "J":
                raise ValueError
        except ValueError:
            raise RotationalStateError("Rotational states must start with" ' "J="')

        state_str = v.strip()

        self.J = None
        if state_str not in ("*", "**", "***"):
            try:
                components = Jstr.parseString(state_str)
            except pp.ParseException:
                raise RotationalStateError(
                    "Invalid rotational state value" " syntax: {0}".format(state_str)
                )

            if "integer" in components:
                self.J = int(components["integer"][0])
            elif "fraction_half" in components:
                self.J = int(components["fraction_half"][0]) / 2
            elif "decimal_half" in components:
                self.J = float(components["decimal_half"][0]) + 0.5
                state_str = "{}/2".format(int(2 * self.J))
            else:
                raise RotationalStateError(
                    "Invalid rotational state value" " syntax: {0}".format(state_str)
                )

            if self.J is not None:
                self._validate_j()

        self.state_str = "J={}".format(state_str)

    def _validate_j(self):
        if self.J % 0.5 != 0:
            raise RotationalStateError(
                "Invalid rotational state value: {}."
                "Must be a multiple of 1/2.".format(self.state_str)
            )

    def __repr__(self):
        return self.state_str
